////////////////////////////////////////////////////////////////////////////////
// Direct Volume Control
// 
// -- AUTHORS --
// * Messer      http://www.dago.pmp.com.pl
// * hartwork    http://www.hartwork.org
////////////////////////////////////////////////////////////////////////////////


#include "foobar2000/SDK/foobar2000.h"
#include <math.h>
#include <mmsystem.h>
#include <process.h>

#include "resource.h"


#define MIXER_ACTION_UP_DB_01		0
#define MIXER_ACTION_UP_DB_10		1
#define MIXER_ACTION_UP_PERC_05		2
#define MIXER_ACTION_UP_PERC_10		3
#define MIXER_ACTION_DOWN_DB_01		4
#define MIXER_ACTION_DOWN_DB_10		5
#define MIXER_ACTION_DOWN_PERC_05	6
#define MIXER_ACTION_DOWN_PERC_10	7
#define MIXER_ACTION_FADE_FULL		8
#define MIXER_ACTION_FADE_ZERO		9
#define MIXER_ACTION_MUTE			10
#define MIXER_ACTION_RESET_INIT		11
#define MIXER_ACTION_FADE_STOP		12
#define MIXER_ACTION_FADE_CLOSE		13

#define MIXER_ACTIONS 14	// MIXER_ACTION_FADE_CLOSE + 1


#define PSEUDO_UNIQUE_BAND_ID		31		// Just some NOT TOO BIG number

#define CONTROL_ID_MENU_ITEM_VOLUME	314		// Just some NOT TOO BIG number
#define CONTROL_ID_MENU_ITEM_HIDE	3		// DO NOT CHANGE!

#define PSEUDO_UNIQUE_TIMER_ID		314159


// Hardcoded settings
#define	FADE_STEPS_PER_SECOND		50		// Bigger is smoother
#define FADE_TIME_MIN 250
#define FADE_TIME_MAX 5000

#define MENU_INDEX_VOLUME			5


static cfg_int g_bIntegrate(		"g_bIntegrate",		FALSE	);
static cfg_int g_bDSPInstead(		"g_bDSPInstead",	FALSE	);
static cfg_int g_iFadeTimeFull(		"g_iFadeTimeFull",	500		);
static cfg_int g_iFadeTimeZero(		"g_iFadeTimeZero",	1700	);
static cfg_int g_bResetOnQuit(		"g_bResetOnQuit",	FALSE	);



DWORD dwInitVol = 0;


HWND wnd_config = NULL;
HWND wnd_fadein = NULL;
HWND wnd_fadeout = NULL;

WNDPROC wndproc_rebar_old = NULL;
WNDPROC wndproc_progress_old = NULL;
WNDPROC wndproc_fadein_old = NULL;
WNDPROC wndproc_fadeout_old = NULL;


//     0 -> -52 db -> -5200
// 65535 ->   0 db ->     0
double	fM = 5200.0 / 65535.0;
int		iN = -5200;


bool bDown = false;


HWND wnd_progress_tooltip =NULL;
TOOLINFO ti_progress;


// Everything accessed by the hooks needs to be shared!
#pragma data_seg( ".SHARED" )

	static cfg_int g_bBandShow(			"g_bBandShow",		TRUE	);
	static cfg_int g_iBandIndex(		"g_iBandIndex",		-1		);
	static cfg_int g_iBandWidth(		"g_iBandWidth",		-1		);
	static cfg_int g_bBandBreak(		"g_bBandBreak",		FALSE	);
	static cfg_int g_bGripperAlways(	"g_bGripperAlways",	TRUE	);
	static cfg_int g_bBandLabel(		"g_bBandLabel",		TRUE	);

	HHOOK hCBT			= NULL;
	HHOOK hMouse		= NULL;

	bool bWaitingMenu	= false;
	bool bWaitingRebar	= false;
	bool bInsertBand	= false;
	bool bAdded			= false;
	bool bNaturalDeath	= false;

	WORD iLastControl	= 0;

	HWND wnd_rebar		= NULL;
	HWND wnd_progress	= NULL;
	HWND wnd_popup		= NULL;

	WNDPROC wndproc_popup_old = NULL;


#pragma data_seg()
#pragma comment( linker, "/section:.SHARED,rws" )



enum FADE_DIRECTION
{
	FADE_FULL,
	FADE_ZERO
};

class mixer
{
private:
	MMRESULT result;
	HMIXER hMixer;
	MIXERLINE ml;
	MIXERLINECONTROLS mlc;
	MIXERCONTROL mc, mc_mute;
	bool bError;

	MIXERCONTROLDETAILS mcd_vol;
	MIXERCONTROLDETAILS mcd_mute;

public:
	mixer( HWND hCallback = NULL )
	{
		// TODO more to init?

		mc.cbStruct			= sizeof( MIXERCONTROL );
		mc_mute.cbStruct	= sizeof( MIXERCONTROL );
		bError = false;


		if( hCallback != NULL )
		{
			result = mixerOpen( &hMixer, 0, ( DWORD )hCallback, 0, MIXER_OBJECTF_HMIXER | CALLBACK_WINDOW );
		}
		else
		{
			result = mixerOpen( &hMixer, 0, 0, 0, MIXER_OBJECTF_HMIXER );
		}

		if( MMSYSERR_NOERROR == result )
		{
			ml.cbStruct			= sizeof( MIXERLINE );
			ml.dwComponentType	= MIXERLINE_COMPONENTTYPE_DST_SPEAKERS;

			result =	mixerGetLineInfo(
							( HMIXEROBJ )hMixer,
							&ml,
							MIXER_GETLINEINFOF_COMPONENTTYPE
						);

			if( MMSYSERR_NOERROR == result )
			{
				// MIXERCONTROL for volume
				mlc.cbStruct		= sizeof( MIXERLINECONTROLS );
				mlc.dwLineID		= ml.dwLineID;
				mlc.dwControlType	= MIXERCONTROL_CONTROLTYPE_VOLUME;
				mlc.cControls		= 1;
				mlc.pamxctrl		= &mc;
				mlc.cbmxctrl		= sizeof( MIXERCONTROL );

				result =	mixerGetLineControls(
								( HMIXEROBJ )hMixer,
								&mlc,
								MIXER_GETLINECONTROLSF_ONEBYTYPE
							);

				if( MMSYSERR_NOERROR != result )
				{
					bError = true;
				}
				else
				{
					mcd_vol.cbStruct	= sizeof( MIXERCONTROLDETAILS );
					mcd_vol.hwndOwner	= 0;
					mcd_vol.dwControlID	= mc.dwControlID;
					// .paDetails is set later
					mcd_vol.cbDetails	= sizeof( MIXERCONTROLDETAILS_UNSIGNED );
					mcd_vol.cChannels	= 1;
				}

				// MIXERCONTROL for mute
				mlc.cbStruct		= sizeof( MIXERLINECONTROLS );
				mlc.dwLineID		= ml.dwLineID;
				mlc.dwControlType	= MIXERCONTROL_CONTROLTYPE_MUTE;
				mlc.cControls		= 1;
				mlc.pamxctrl		= &mc_mute;
				mlc.cbmxctrl		= sizeof( MIXERCONTROL );
				
				result =	mixerGetLineControls(
								( HMIXEROBJ )hMixer,
								&mlc,
								MIXER_GETLINECONTROLSF_ONEBYTYPE
							);

				if( MMSYSERR_NOERROR != result )
				{
					bError = true;
				}
				else
				{
					mcd_mute.cbStruct		= sizeof( MIXERCONTROLDETAILS );
					mcd_mute.hwndOwner		= 0;
					mcd_mute.dwControlID	= mc_mute.dwControlID;
					// .paDetails is set later
					mcd_mute.cbDetails		= sizeof( MIXERCONTROLDETAILS_UNSIGNED );
					mcd_mute.cChannels		= 1;
				}
			}
			else
			{
				bError = true;
			}
		}
		else
		{
			hMixer = NULL;
			bError = true;

			console::error( "could not open mixer device" );
			console::popup();
		}
	}

	~mixer()
	{
		if( hMixer ) mixerClose( hMixer );
	}

	DWORD getVolumeRAW()
	{
		if( bError ) return 0;

		// if we don't run these two each time we get in trouble
		// on win2k. don't ask me why...

		MIXERCONTROLDETAILS_UNSIGNED mcdu_vol = { 0 };
		mcd_vol.paDetails = &mcdu_vol;

		result =	mixerGetControlDetails(
						( HMIXEROBJ )hMixer,
						&mcd_vol,
						MIXER_GETCONTROLDETAILSF_VALUE // | MIXER_OBJECTF_HMIXER
					);

		if( MMSYSERR_NOERROR != result )
		{
			bError = true;
			mcdu_vol.dwValue = -1;

			switch( result )
			{
				case MIXERR_INVALCONTROL:
				{
					console::error( "MIXERR_INVALCONTROL" );
					break;
				}
				case MMSYSERR_BADDEVICEID:
				{
					console::error( "MMSYSERR_BADDEVICEID" );
					break;
				}
				case MMSYSERR_INVALFLAG:
				{
					console::error( "MMSYSERR_INVALFLAG" );
					break;
				}
				case MMSYSERR_INVALHANDLE:
				{
					console::error( "MMSYSERR_INVALHANDLE" );
					break;
				}
				case MMSYSERR_INVALPARAM:
				{
					console::error( "MMSYSERR_INVALPARAM" );
					break;
				}
				case MMSYSERR_NODRIVER: 
				{
					console::error( "MMSYSERR_NODRIVER" );
					break;
				}
				default:
				{
					console::error( "unknown error" );
					break;
				}
			}
			console::popup();
		}
		return mcdu_vol.dwValue;
	}

	static double conv_RAW_to_Percent( DWORD dw )
	{
		return ( dw * 100.0f / 65535.0f );
	}

	static DWORD conv_RAW_to_DB( DWORD dw )
	{
		return ( DWORD )( 200 * log10( ( double )dw ) );
	}

	double getVolumePercent()
	{
		return conv_RAW_to_Percent( getVolumeRAW() );
	}

	DWORD getVolumeDB()
	{
		return conv_RAW_to_DB( getVolumeRAW() );
	}

	void setVolumeRAW( DWORD dwVolume )
	{
		// if we don't run these two each time we get in trouble
		// on win2k. don't ask me why...

		MIXERCONTROLDETAILS_UNSIGNED mcdu_vol = { 0 };
		mcd_vol.paDetails = &mcdu_vol;


		// The conversion to long is important
		// to keep out negative numbers!

		mcdu_vol.dwValue = __max( __min( ( long )dwVolume, 65535 ), 0 );

		result =	mixerSetControlDetails(
						( HMIXEROBJ )hMixer,
						&mcd_vol, // &mcd,
					    MIXER_SETCONTROLDETAILSF_VALUE
					);
		
		if( MMSYSERR_NOERROR != result )
		{
			bError = true;

			switch( result )
			{
				case MIXERR_INVALCONTROL:
				{
					console::error( "MIXERR_INVALCONTROL" );
					break;
				}
				case MMSYSERR_BADDEVICEID:
				{
					console::error( "MMSYSERR_BADDEVICEID" );
					break;
				}
				case MMSYSERR_INVALFLAG:
				{
					console::error( "MMSYSERR_INVALFLAG" );
					break;
				}
				case MMSYSERR_INVALHANDLE:
				{
					console::error( "MMSYSERR_INVALHANDLE" );
					break;
				}
				case MMSYSERR_INVALPARAM:
				{
					console::error( "MMSYSERR_INVALPARAM" );
					break;
				}
				case MMSYSERR_NODRIVER: 
				{
					console::error( "MMSYSERR_NODRIVER" );
					break;
				}
				default:
				{
					console::error( "unknown error" );
					break;
				}
			}
			console::popup();
		}
	}

	static DWORD conv_Percent_to_RAW( double f )
	{
		return ( DWORD )( 65535 * f / 100 );
	}

	static DWORD conv_DB_to_RAW( DWORD dw )
	{
		return ( DWORD )pow( 10.0, ( double )( dw ) / 200.0 );
	}

	void setVolumePercent( double fPercent )
	{
		setVolumeRAW( conv_Percent_to_RAW( fPercent ) );
	}

	void setVolumeDB( DWORD dwDB )
	{
		setVolumeRAW( conv_DB_to_RAW( dwDB ) );
	}

	void setMute( bool mute )
	{
		// if we don't run these two each time we get in trouble
		// on win2k. don't ask me why...
		MIXERCONTROLDETAILS_UNSIGNED mcdu_mute = { 0 };
		mcd_mute.paDetails = &mcdu_mute;
		
		mcdu_mute.dwValue = mute ? 0 : 1;

		result =	mixerSetControlDetails(
						( HMIXEROBJ )hMixer,
						&mcd_mute,
						MIXER_SETCONTROLDETAILSF_VALUE
					);

		if( MMSYSERR_NOERROR != result )
		{
			bError = true;

			switch( result )
			{
				case MIXERR_INVALCONTROL:
				{
					console::error( "MIXERR_INVALCONTROL" );
					break;
				}
				case MMSYSERR_BADDEVICEID:
				{
					console::error( "MMSYSERR_BADDEVICEID" );
					break;
				}
				case MMSYSERR_INVALFLAG:
				{
					console::error( "MMSYSERR_INVALFLAG" );
					break;
				}
				case MMSYSERR_INVALHANDLE:
				{
					console::error( "MMSYSERR_INVALHANDLE" );
					break;
				}
				case MMSYSERR_INVALPARAM:
				{
					console::error( "MMSYSERR_INVALPARAM" );
					break;
				}
				case MMSYSERR_NODRIVER: 
				{
					console::error( "MMSYSERR_NODRIVER" );
					break;
				}
				default:
				{
					console::error( "unknown error" );
					break;
				}
			}
			console::popup();
		}
	}


	bool getMute()
	{
		if( bError ) return 0;

		// if we don't run these two each time we get in trouble
		// on win2k. don't ask me why...
		MIXERCONTROLDETAILS_UNSIGNED mcdu_mute = { 0 };
		mcd_mute.paDetails = &mcdu_mute;

		result =	mixerGetControlDetails(
						( HMIXEROBJ )hMixer,
						&mcd_mute,
						MIXER_GETCONTROLDETAILSF_VALUE
					);

		bool res = false;
	    if( MMSYSERR_NOERROR == result )
		{
			res = ( mcdu_mute.dwValue > 0 ) ? true : false;
		}
		else
		{
			bError = true;
			res = false;

			switch( result )
			{
				case MIXERR_INVALCONTROL:
				{
					console::error( "MIXERR_INVALCONTROL" );
					break;
				}
				case MMSYSERR_BADDEVICEID:
				{
					console::error( "MMSYSERR_BADDEVICEID" );
					break;
				}
				case MMSYSERR_INVALFLAG:
				{
					console::error( "MMSYSERR_INVALFLAG" );
					break;
				}
				case MMSYSERR_INVALHANDLE:
				{
					console::error( "MMSYSERR_INVALHANDLE" );
					break;
				}
				case MMSYSERR_INVALPARAM:
				{
					console::error( "MMSYSERR_INVALPARAM" );
					break;
				}
				case MMSYSERR_NODRIVER: 
				{
					console::error( "MMSYSERR_NODRIVER" );
					break;
				}
				default:
				{
					console::error( "unknown error" );
					break;
				}
			}
			console::popup();
		}
		return !res;
	}

	// Functions for private member access
	HMIXER getMixerHandle() { return hMixer; }
	bool getError() { return bError; }
	void setError() { bError = true; }
	DWORD getVolumeControlID() { return mc.dwControlID; }
	DWORD getMuteControlID() { return mc_mute.dwControlID; }
};

mixer * m = NULL;


/******************************************************************************************************
**  hookproc_mouse                                                                                   **
******************************************************************************************************/

LRESULT CALLBACK hookproc_mouse( int code, WPARAM wp, LPARAM lp )
{
	if( code >= 0 )
	{
		if( HC_ACTION == code )
		{
			if( ( WM_LBUTTONUP == wp ) || ( WM_RBUTTONUP == wp ) )
			{
				// TODO: Is "noright" on?

				MOUSEHOOKSTRUCT * mhs = ( MOUSEHOOKSTRUCT * )lp;
				if( mhs->hwnd == wnd_rebar )
				{
					bNaturalDeath = true;
				}
			}
		}
	}
	return CallNextHookEx( hMouse, code, wp, lp );
}


/******************************************************************************************************
**  wndproc_popup                                                                                    **
******************************************************************************************************/

LRESULT CALLBACK wndproc_popup( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	switch( message )
	{
		case WM_NCCREATE:	// first message ever
		{
			// Menu windows don't get much mouse info
			// so we need a hook to detect a click

			hMouse = SetWindowsHookEx( WH_MOUSE, &hookproc_mouse, core_api::get_my_instance(), GetCurrentThreadId() );
			bNaturalDeath = false;

			break;
		}
		case 482:			// Menu is present now
		{
			// We emulate a down-keystroke which will
			// produce a WM_MENUSELECT.
			// Latter will give us the needed menu handle!

			SendMessage( hwnd, WM_KEYDOWN, VK_DOWN, 0 );

			break;
		}

/*
		case 485:			// Item was selected
		{
			// wParam is the new index or -1 if
			// the cursor is moving outside the window

			// We don't check anything here
			// cause we're checking on WM_MENUSELECT

			break;
		}
*/

		case WM_DESTROY:
		{
			// We don't need the hook anymore
			if( hMouse != NULL ) UnhookWindowsHookEx( hMouse );

			bAdded = false;

			// Was an item chosen
			if( bNaturalDeath )
			{
				switch( iLastControl )
				{
					case CONTROL_ID_MENU_ITEM_VOLUME:
					{
						// Get band visibilty
						REBARBANDINFO rbbi_get;
						ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
						rbbi_get.cbSize	= sizeof( REBARBANDINFO );
						rbbi_get.fMask	= RBBIM_STYLE;

						int iCurIndex = SendMessage( wnd_rebar, RB_IDTOINDEX, ( WPARAM )( UINT )PSEUDO_UNIQUE_BAND_ID, 0 );
						SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )iCurIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

						
						// Show/hide out band
						SendMessage( wnd_rebar, RB_SHOWBAND, ( WPARAM )( UINT )iCurIndex, ( LPARAM )( bool )( rbbi_get.fStyle & RBBS_HIDDEN ) );

						break;
					}
					case CONTROL_ID_MENU_ITEM_HIDE:
					{
						// Get band visibilty/size/break
						REBARBANDINFO rbbi_get;
						ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
						rbbi_get.cbSize	= sizeof( REBARBANDINFO );
						rbbi_get.fMask	= RBBIM_STYLE | RBBIM_SIZE;
	
						int iCurIndex = SendMessage( wnd_rebar, RB_IDTOINDEX, ( WPARAM )( UINT )PSEUDO_UNIQUE_BAND_ID, 0 );
						SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )iCurIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

						// Remove band (or trouble)
						SendMessage( wnd_rebar, RB_DELETEBAND, ( WPARAM )iCurIndex, 0 );
		
						// Save band settings
						g_iBandIndex	= iCurIndex;
						g_bBandShow		= !( rbbi_get.fStyle & RBBS_HIDDEN );
						g_iBandWidth	= rbbi_get.cx;
						g_bBandBreak	= ( bool )( rbbi_get.fStyle & RBBS_BREAK );


						// Change progress window to NON-CHILD
						SetWindowLong( wnd_progress, GWL_STYLE, 0 );
						SetParent( wnd_progress, NULL );


						// We are searching again
						bWaitingRebar = true;

						break;
					}
				}
			}
			break;
		}
		case WM_KEYDOWN:
		{
			if( VK_RETURN == wp )
			{
				bNaturalDeath = true;
			}
			break;
		}
	}
	return CallWindowProc( wndproc_popup_old, hwnd, message, wp, lp );
}


/******************************************************************************************************
**  wndproc_rebar                                                                                    **
******************************************************************************************************/

LRESULT CALLBACK wndproc_rebar( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	static bool bOthersBefore = false;

	switch( message )
	{
		case WM_CREATE:
		{
			bOthersBefore = false;

			break;
		}
		case WM_CONTEXTMENU:
		{
			if( g_bIntegrate )
			{
				bWaitingMenu = true;
			}

			break;
		}
		case WM_MENUSELECT:
		{
			HMENU menu = ( HMENU )lp;
			if( menu == NULL ) break;

			iLastControl = LOWORD( wp );

			if( IsWindow( wnd_popup ) && !bAdded )
			{
				// Get band visibility
				REBARBANDINFO rbbi_get;
				ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
				rbbi_get.cbSize	= sizeof( REBARBANDINFO );
				rbbi_get.fMask	= RBBIM_STYLE;

				int iCurIndex = SendMessage( wnd_rebar, RB_IDTOINDEX, ( WPARAM )( UINT )PSEUDO_UNIQUE_BAND_ID, 0 );
				SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )iCurIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

				// Modify menu
				MENUITEMINFO mii;
				ZeroMemory( &mii, sizeof( MENUITEMINFO ) );
		        mii.cbSize		= sizeof( MENUITEMINFO );
			    mii.fMask		= MIIM_STATE | MIIM_ID | MIIM_TYPE;
				mii.wID			= CONTROL_ID_MENU_ITEM_VOLUME;
	            mii.fType		= MFT_STRING;
				mii.fState		= ( ( rbbi_get.fStyle & RBBS_HIDDEN ) ? MFS_UNCHECKED : MFS_CHECKED ) | MFS_ENABLED;
				mii.dwTypeData	= "Volume";
				mii.cch			= strlen( mii.dwTypeData );
	            InsertMenuItem( menu, MENU_INDEX_VOLUME, TRUE, &mii );

				// One is enough
				bAdded = true;
			}
			break;
		}
		case RB_INSERTBANDA:
		case RB_INSERTBANDW:
		{
			bOthersBefore = true;
			break;
		}
		case RB_SHOWBAND:
		{
			// Other bands have been inserted before
			if( bInsertBand && bOthersBefore )
			{
				// Change progress to CHILD
				SetParent( wnd_progress, wnd_rebar );
				SetWindowLong( wnd_progress, GWL_STYLE, WS_CHILD );



				// Add band
				REBARBANDINFO rbbi_vol;
				ZeroMemory( &rbbi_vol, sizeof( REBARBANDINFO ) );
				rbbi_vol.cbSize			= sizeof( REBARBANDINFO );
				rbbi_vol.fMask			= RBBIM_CHILD | RBBIM_CHILDSIZE | RBBIM_ID | RBBIM_STYLE |
											( g_bBandLabel ? RBBIM_TEXT : 0 ) |
											( ( g_iBandWidth != -1 ) ? RBBIM_SIZE : 0 );
				rbbi_vol.hwndChild		= wnd_progress;
				rbbi_vol.lpText			= "Volume";
				rbbi_vol.cxMinChild		= 113;
				rbbi_vol.cyMinChild		= 18;
				rbbi_vol.cyChild		= 18;
				rbbi_vol.cyMaxChild 	= 18;
				rbbi_vol.wID			= PSEUDO_UNIQUE_BAND_ID;
				rbbi_vol.cx				= g_iBandWidth;
				rbbi_vol.fStyle			= RBBS_CHILDEDGE | ( g_bGripperAlways ? RBBS_GRIPPERALWAYS : 0 ) |
											( ( g_bBandShow ) ? 0 : RBBS_HIDDEN ) |
											( ( g_bBandBreak ) ? RBBS_BREAK : 0 );

				int iBandCount = 0;
				if( g_iBandIndex > ( iBandCount = SendMessage( wnd_rebar, RB_GETBANDCOUNT, 0, 0 ) ) )
				{
					g_iBandIndex = iBandCount;
				}

				SendMessage( wnd_rebar, RB_INSERTBAND, ( WPARAM )g_iBandIndex, ( LPARAM )( LPREBARBANDINFO )&rbbi_vol );

				if( !g_bBandShow ) ShowWindow( wnd_progress, SW_HIDE );



				// RBBS_GRIPPERALWAYS on/off + check for RBBS_NOGRIPPER
				bool bNoGripper = false;
				iBandCount++;
				for( int i = 0; i < iBandCount; i++ )
				{
					// Exclude volume band
					if( i == g_iBandIndex ) continue;

					// Get band info
					REBARBANDINFO rbbi_get;
					ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
					rbbi_get.cbSize	= sizeof( REBARBANDINFO );
					rbbi_get.fMask	= RBBIM_STYLE;

					SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )i, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

					// Add RBBS_GRIPPERALWAYS where missing
					if( ( bool )g_bGripperAlways ^ ( bool )( rbbi_get.fStyle & RBBS_GRIPPERALWAYS ) )		// xor
					{
						if( g_bGripperAlways )
						{
							rbbi_get.fStyle	|= RBBS_GRIPPERALWAYS;
						}
						else
						{
							rbbi_get.fStyle	-= RBBS_GRIPPERALWAYS;
						}

						// Only the values specified in [.fMask] are overwritten???
						SendMessage( wnd_rebar, RB_SETBANDINFO, ( WPARAM )( UINT )i, ( LPARAM )( REBARBANDINFO * )&rbbi_get );
					}

					// Toolbar locked?
					bNoGripper |= ( bool )( rbbi_get.fStyle & RBBS_NOGRIPPER );
				}



				// Update volume band
				if( bNoGripper )
				{
					rbbi_vol.fStyle |= RBBS_NOGRIPPER;
					SendMessage( wnd_rebar, RB_SETBANDINFO, ( WPARAM )( UINT )g_iBandIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_vol );
				}



				// One band is enough
				bInsertBand = false;
			}
			break;
		}

/*
		case WM_DESTROY:
		{
			// If foo_ui_default is used and foobar is
			// being shutdown we don't get here
			// If foo_ui_default is used and [hide toolbar]
			// was chosen bWaitingRebar must be true

			if( !bWaitingRebar )
			{
				// UI different to foo_ui_default
				// Columns UI for example
			}

			break;
		}
*/

	}
	return CallWindowProc( wndproc_rebar_old, hwnd, message, wp, lp );
}


/******************************************************************************************************
**  wndproc_progress                                                                                 **
******************************************************************************************************/

LRESULT CALLBACK wndproc_progress( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	switch( message )
	{
		case MM_MIXM_CONTROL_CHANGE:
		{
			// Don't update our slider...
			// * when we are sliding
			// * if we work with dsp
			if( bDown || g_bDSPInstead ) break;

			// Is it main volume?
			if
			(
				( wp == ( WPARAM )m->getMixerHandle() ) &&
				( lp == ( LPARAM )m->getVolumeControlID() )
			)
			{
				if( m )
				{
					// Apply current volume
					PostMessage( hwnd, PBM_SETPOS , ( WPARAM )( m->getVolumeRAW() ), 0 );
				}
			}
			break;
		}
		case WM_LBUTTONDOWN:
		{
			// Show tooltip
			RECT r1;
			GetWindowRect( wnd_rebar, &r1 );
			RECT r2;
			int iCurIndex = SendMessage( wnd_rebar, RB_IDTOINDEX, ( WPARAM )( UINT )PSEUDO_UNIQUE_BAND_ID, 0 );
			SendMessage( wnd_rebar, RB_GETRECT, ( WPARAM )( INT )iCurIndex, ( LPARAM )( LPRECT )&r2 );

			SendMessage( wnd_progress_tooltip, TTM_TRACKPOSITION, 0, ( LPARAM )( DWORD )MAKELONG( r1.left + r2.left + 6, r1.top + r2.top + 6 ) );
			SendMessage( wnd_progress_tooltip, TTM_TRACKACTIVATE, true, ( LPARAM )( LPTOOLINFO )&ti_progress );

			
			// Capture mouse
			SetCapture( hwnd );
			bDown = true;

			// NO BREAK!
		}
		case WM_MOUSEMOVE:
		{
			if( !bDown ) break;

			RECT r;
			GetWindowRect( hwnd, &r );

			const int iWidth	= r.right - r.left;
			int iVal			= 65535 * __max( 0, __min( ( short )LOWORD( lp ), iWidth ) ) / iWidth;
			
			// Update slider
			PostMessage( hwnd, PBM_SETPOS , ( WPARAM )( iVal ), 0 );

			// Update volume and tooltip
			char * szBuffer = new char[ 8 ];

			if( g_bDSPInstead )
			{
				iVal = ( int )( fM * iVal ) + iN;

				float fDB = iVal * 0.01f;
				sprintf( szBuffer, "%.2fdB", fDB );
	
				play_control::get()->set_volume( iVal );
			}
			else
			{
				float fPercent = ( float )iVal * 100 / 65535;
				sprintf( szBuffer, "%.1f%%", fPercent );

				m->setVolumeRAW( iVal );
			}

			// Quick crash fix 2006-04-20
			// strcpy( ti_progress.lpszText, szBuffer );
			ti_progress.lpszText = szBuffer;
			SendMessage( wnd_progress_tooltip, TTM_UPDATETIPTEXT, 0, ( LPARAM )( LPTOOLINFO )&ti_progress );
			delete szBuffer;

			break;
		}
		case WM_LBUTTONUP:
		case WM_NCLBUTTONUP:
		{
			// Hide tooltip
			SendMessage( wnd_progress_tooltip, TTM_TRACKACTIVATE, false, ( LPARAM )( LPTOOLINFO )&ti_progress );

			// Release capture
			bDown = false;
			ReleaseCapture();

			break;
		}
	}
	return CallWindowProc( wndproc_progress_old, hwnd, message, wp, lp );
}


/******************************************************************************************************
**  FADE                                                                                             **
******************************************************************************************************/

namespace FADE
{
	bool bTimerProcRunning	= false;
	UINT iCounter			= 0;
	bool bActive			= false;
	bool bStopAfter			= false;
	bool bCloseAfter		= false;
	FADE_DIRECTION dir		= FADE_ZERO;
	double fStep			= 0.0f;
	UINT delay_per_step		= 0;
	double x				= 0.0f;
	DWORD dwVolumeBefore	= 65535;

	const int factor = 3;

	double f( DWORD x )		// Linear fading sucks
	{
		// return ( double )( atan( x / 65535.0f ) * 65535.0f / atan( 1.0f ) );
		return ( double )( atan( x * factor / 65535.0f ) * 65535.0f / atan( ( float )factor ) );
		
	}
	
	double f_1( double x )	// Inverse fade function
	{
		// return ( double )tan( x * atan( 1.0f ) / 65535.0f ) * 65535.0f;
		return ( double )tan( x * atan( ( float )factor ) / 65535.0f ) * 65535.0f / factor;
	}

	VOID CALLBACK fadeproc( HWND hwnd, UINT message, UINT idEvent, DWORD dwTime )
	{
		if( bTimerProcRunning ) return;
		bTimerProcRunning = true;

		iCounter--;

		if( iCounter )
		{
			x += fStep;
			m->setVolumeRAW( ( DWORD )f( x ) );
		}
		else
		{
			// Last fade step
			if( FADE_FULL == dir )
			{
				m->setVolumeRAW( 65535 );
			}
			else // FADE_ZERO
			{
				m->setVolumeRAW( 0 );
				if( bStopAfter )
				{
					// Stop playback
					if( play_control::get()->is_playing() && !play_control::get()->is_paused() )
					{
						play_control::get()->play_stop();
					}
						
					// Reset volume
					m->setVolumeRAW( dwVolumeBefore );

					
					// Close foobar
					if( bCloseAfter )
					{
						PostMessage( core_api::get_main_window(), WM_CLOSE, 0, 0 );
					}
				}
			}
			bActive = false;
			KillTimer( wnd_progress, PSEUDO_UNIQUE_TIMER_ID );
		}
		bTimerProcRunning = false;
	}

	void doFade( FADE_DIRECTION dir_new, bool stop_after_new, bool close_after_new )
	{
		// Already fading?
		if( bActive )
		{


/******************************************************************************************************
**          Modify active fade                                                                       **
******************************************************************************************************/
			
			if( FADE_FULL == dir_new )
			{
				// Same direction as before?
				if( dir_new == dir )
				{
					return;
				}


/******************************************************************************************************
**              ZERO before -> now FULL                                                              **
******************************************************************************************************/

				DWORD y = ( DWORD )f( x );

				// Already FULL?
				if( 65535 == y )
				{
					return;
				}

				dir = dir_new;

				delay_per_step = g_iFadeTimeFull / FADE_STEPS_PER_SECOND;
				iCounter = g_iFadeTimeFull * FADE_STEPS_PER_SECOND / 1000;
				fStep = ( 65535.0f - x ) / iCounter;

				bActive      = true;
				bStopAfter   = false;
				bCloseAfter  = false;
			}
			else // FADE_ZERO
			{
				// Same direction as before?
				if( dir_new == dir )
				{
					bStopAfter   = stop_after_new;
					bCloseAfter  = close_after_new;

					return;	
				}


/******************************************************************************************************
**              FULL before -> now ZERO                                                              **
******************************************************************************************************/

				DWORD y = ( DWORD )f( x );

				// Already ZERO?
				if( 0 == y )
				{
					if( stop_after_new )
					{
						// Stop playback
						if( play_control::get()->is_playing() && !play_control::get()->is_paused() )
						{
							play_control::get()->play_stop();
						}
						

						// Reset volume
						m->setVolumeRAW( dwVolumeBefore );


						// Close foobar
						if( close_after_new )
						{
							PostMessage( core_api::get_main_window(), WM_CLOSE, 0, 0 );
						}
					}

					return;
				}

				dir = dir_new;

				delay_per_step	= g_iFadeTimeZero / FADE_STEPS_PER_SECOND;
				iCounter		= g_iFadeTimeZero * FADE_STEPS_PER_SECOND / 1000;
				fStep			= ( -1 )*( x / iCounter );

				bActive			= true;
				bStopAfter		= stop_after_new;
				bCloseAfter		= close_after_new;

				dwVolumeBefore = 65535;
			}

			// Replace timer, call first time
			KillTimer( wnd_progress, PSEUDO_UNIQUE_TIMER_ID );
			fadeproc( wnd_progress, WM_TIMER, PSEUDO_UNIQUE_TIMER_ID, 0 );
			SetTimer( wnd_progress, PSEUDO_UNIQUE_TIMER_ID, delay_per_step, fadeproc );
		}
		else
		{


/******************************************************************************************************
**          Start new fade                                                                           **
******************************************************************************************************/

			DWORD y = m->getVolumeRAW();
			if( FADE_FULL == dir_new )
			{
				// Already FULL?
				if( 65535 == y )
				{
					return;
				}


/******************************************************************************************************
**              new FULL                                                                             **
******************************************************************************************************/

				x = f_1( ( double )y );

				dir = dir_new;

				delay_per_step	= g_iFadeTimeFull / FADE_STEPS_PER_SECOND;
				iCounter		= g_iFadeTimeFull * FADE_STEPS_PER_SECOND / 1000;
				fStep			= ( 65535.0f - x ) / iCounter;

				bActive		 = true;
				bStopAfter	 = false;
				bCloseAfter  = false;
			}
			else // FADE_ZERO
			{
				// Already ZERO?
				if( 0 == y )
				{
					if( stop_after_new )
					{	
						// Stop playback
						if( play_control::get()->is_playing() && !play_control::get()->is_paused() )
						{
							play_control::get()->play_stop();
						}

						
						// Nothing to reset this times


						// Close foobar
						if( close_after_new )
						{
							PostMessage( core_api::get_main_window(), WM_CLOSE, 0, 0 );
						}
					}

					return;
				}


/******************************************************************************************************
**              new ZERO                                                                             **
******************************************************************************************************/

				x = f_1( ( double )y );

				dir = dir_new;

				delay_per_step	= g_iFadeTimeZero / FADE_STEPS_PER_SECOND;
				iCounter		= g_iFadeTimeZero * FADE_STEPS_PER_SECOND / 1000;
				fStep			= ( -1 )*( x / iCounter );

				bActive			= true;
				bStopAfter		= stop_after_new;
				bCloseAfter		= close_after_new;

				dwVolumeBefore	= y;
			}

			// Create new timer
			fadeproc( wnd_progress, WM_TIMER, PSEUDO_UNIQUE_TIMER_ID, 0 );
			SetTimer( wnd_progress, PSEUDO_UNIQUE_TIMER_ID, delay_per_step, fadeproc );
		}
	}
};


/******************************************************************************************************
**  hookproc_cbt                                                                                     **
******************************************************************************************************/

LRESULT CALLBACK hookproc_cbt( int code, WPARAM wp, LPARAM lp )
{
	if( code == HCBT_CREATEWND )
	{
		// MSDN says WE CANNOT TRUST "CBT_CREATEWND"
		// so we use only the window handle
		// and get the class using "GetClassName". (-> Q106079)

		if( bWaitingMenu )
		{
			HWND hwnd = ( HWND )wp;
			char szClass[ 7 ] = "";
			GetClassName( hwnd, szClass, 7 );

			// Popup menu?
			if( !strcmp( szClass, "#32768" ) )
			{
				// We will simulate a mousemove to raise WM_MENUSELECT
				// which gives us the HMENU handle...

				wnd_popup = hwnd;

				// Exchange wndproc POPUP
				wndproc_popup_old = ( WNDPROC )GetWindowLong( hwnd, GWL_WNDPROC );
				if( wndproc_popup_old != NULL )
				{
					SetWindowLong( hwnd, GWL_WNDPROC, ( LONG )wndproc_popup );
				}

				bWaitingMenu = false;
			}
		}
		else if( bWaitingRebar )
		{
			HWND hwnd = ( HWND )wp;
			char szClass[ 30 ] = "";
			GetClassName( hwnd, szClass, 30 );

			// Rebar control?
			if( !strcmp( szClass, REBARCLASSNAME ) )
			{
				wnd_rebar = hwnd;

				// Exchange wndproc REBAR
				wndproc_rebar_old = ( WNDPROC )GetWindowLong( wnd_rebar, GWL_WNDPROC );
				if( wndproc_rebar_old != NULL )
				{
					SetWindowLong( wnd_rebar, GWL_WNDPROC, ( LONG )wndproc_rebar );
				}

				// Allow band insertion
				bInsertBand = true;

				// Stop searching
				bWaitingRebar = false;
			}
		}
	}	
	return CallNextHookEx( hCBT, code, wp, lp );
}

class mixer_init : public initquit
{
	virtual void on_init()
	{
		// Ensure dll is loaded
		INITCOMMONCONTROLSEX icce;
		icce.dwSize = sizeof( INITCOMMONCONTROLSEX );
		icce.dwICC =	ICC_COOL_CLASSES |		// Rebar
						ICC_PROGRESS_CLASS |	// Progress
						ICC_BAR_CLASSES;		// Tooltips

		InitCommonControlsEx( &icce );


		// Check for toolbar
		wnd_rebar = FindWindowEx( core_api::get_main_window(), NULL, REBARCLASSNAME, NULL );
		if( !( bWaitingRebar = ( wnd_rebar == NULL ) ) )
		{
			// Rebar control found

			// Exchange wndproc REBAR
			wndproc_rebar_old = ( WNDPROC )GetWindowLong( wnd_rebar, GWL_WNDPROC );
			if( wndproc_rebar_old != NULL )
			{
				SetWindowLong( wnd_rebar, GWL_WNDPROC, ( LONG )wndproc_rebar );
			}

			// Create progressbar as CHILD
			wnd_progress =	CreateWindowEx
							(
								0,
								PROGRESS_CLASS, 
								NULL,
								WS_CHILD,
								0, 0, 113, 18,			// TODO restore saved width
								wnd_rebar,				// PARENT ! 
								NULL, 
								core_api::get_my_instance(), 
								NULL
							);


			if( g_bIntegrate )
			{
				// Add band
				REBARBANDINFO rbbi_vol;
				ZeroMemory( &rbbi_vol, sizeof( REBARBANDINFO ) );
				rbbi_vol.cbSize			= sizeof( REBARBANDINFO );
				rbbi_vol.fMask			= RBBIM_CHILD | RBBIM_CHILDSIZE | RBBIM_ID | RBBIM_STYLE |
											( g_bBandLabel ? RBBIM_TEXT : 0 ) |
											( ( g_iBandWidth != -1 ) ? RBBIM_SIZE : 0 );
				rbbi_vol.hwndChild		= wnd_progress;
				rbbi_vol.lpText			= "Volume";
				rbbi_vol.cxMinChild		= 113;
				rbbi_vol.cyMinChild		= 18;
				rbbi_vol.cyChild		= 18;
				rbbi_vol.cyMaxChild 	= 18;
				rbbi_vol.wID			= PSEUDO_UNIQUE_BAND_ID;
				rbbi_vol.cx				= g_iBandWidth;
				rbbi_vol.fStyle			= RBBS_CHILDEDGE | ( g_bGripperAlways ? RBBS_GRIPPERALWAYS : 0 ) |
											( ( g_bBandShow ) ? 0 : RBBS_HIDDEN ) |
											( ( g_bBandBreak ) ? RBBS_BREAK : 0 );

				int iBandCount = 0;
				if( g_iBandIndex > ( iBandCount = SendMessage( wnd_rebar, RB_GETBANDCOUNT, 0, 0 ) ) )
				{
					g_iBandIndex = iBandCount;
				}

				SendMessage( wnd_rebar, RB_INSERTBAND, ( WPARAM )g_iBandIndex, ( LPARAM )( LPREBARBANDINFO )&rbbi_vol );

				if( !g_bBandShow ) ShowWindow( wnd_progress, SW_HIDE );


		
				// RBBS_GRIPPERALWAYS on/off + check for RBBS_NOGRIPPER
				bool bNoGripper = false;
				iBandCount++;
				for( int i = 0; i < iBandCount; i++ )
				{
					// Exclude volume band
					if( i == g_iBandIndex ) continue;

					// Get band info
					REBARBANDINFO rbbi_get;
					ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
					rbbi_get.cbSize	= sizeof( REBARBANDINFO );
					rbbi_get.fMask	= RBBIM_STYLE;

					SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )i, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

					// Add RBBS_GRIPPERALWAYS where missing
					if( ( bool )g_bGripperAlways ^ ( bool )( rbbi_get.fStyle & RBBS_GRIPPERALWAYS ) )		// xor
					{
						if( g_bGripperAlways )
						{
							rbbi_get.fStyle	|= RBBS_GRIPPERALWAYS;
						}
						else
						{
							rbbi_get.fStyle	-= RBBS_GRIPPERALWAYS;
						}

						// Only the values specified in [.fMask] are overwritten???
						SendMessage( wnd_rebar, RB_SETBANDINFO, ( WPARAM )( UINT )i, ( LPARAM )( REBARBANDINFO * )&rbbi_get );
					}

					// Toolbar locked?
					bNoGripper |= ( bool )( rbbi_get.fStyle & RBBS_NOGRIPPER );
				}

				// Remove volume band gripper
				if( bNoGripper )
				{
					rbbi_vol.fStyle |= RBBS_NOGRIPPER;
					SendMessage( wnd_rebar, RB_SETBANDINFO, ( WPARAM )( UINT )g_iBandIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_vol );
				}
			}
		}
		else
		{
			// Rebar control NOT found

			// Create progressbar as NON-CHILD
			InitCommonControls();
			wnd_progress =	CreateWindowEx
							(
								0,
								PROGRESS_CLASS, 
								NULL,
								0,
								0, 0, 113, 18,				// TODO restore saved width
								NULL,						// PARENT !
								NULL, 
								core_api::get_my_instance(), 
								NULL
							);
		}

		// Setup progressbar
		SendMessage(wnd_progress, PBM_SETRANGE, 0, MAKELPARAM( 0, 65535 ) ); 
	    SendMessage(wnd_progress, PBM_SETSTEP , ( WPARAM )4096, 0 );

		// Exchange wndproc PROGRESS
		wndproc_progress_old = ( WNDPROC )GetWindowLong( wnd_progress, GWL_WNDPROC );
		if( wndproc_progress_old != NULL )
		{
			SetWindowLong( wnd_progress, GWL_WNDPROC, ( LONG )wndproc_progress );
		}

		// Create mixer w/ callback to progress window
		if( !m ) m = new mixer( wnd_progress );

		// Save initial volume
		dwInitVol = m->getVolumeRAW();

		// Apply current volume to progress bar
		if( g_bDSPInstead )
		{
			int iVal = ( int )( ( play_control::get()->get_volume() - iN ) / fM ) + 1;
			iVal = __max( 0, __min( iVal, 65535 ) );
			PostMessage( wnd_progress, PBM_SETPOS , ( WPARAM )iVal, 0 );
		}
		else
		{
			SendMessage( wnd_progress, PBM_SETPOS , ( WPARAM )dwInitVol, 0 );
		}


		// Create tooltip window
		wnd_progress_tooltip =	CreateWindowEx
								(
									WS_EX_TOPMOST,
									TOOLTIPS_CLASS,
									NULL,
									TTS_NOPREFIX | TTS_ALWAYSTIP,		
									CW_USEDEFAULT,
									CW_USEDEFAULT,
									CW_USEDEFAULT,
									CW_USEDEFAULT,
									NULL,
									NULL,
									NULL,
									NULL
								);

		ti_progress.cbSize		= sizeof( TOOLINFO );
		ti_progress.uFlags		= TTF_TRACK;
		ti_progress.hwnd		= NULL;
		ti_progress.hinst		= NULL;
		ti_progress.uId			= 0;
		ti_progress.lpszText	= " ";		// Or trouble
		ZeroMemory( &ti_progress.rect, sizeof( RECT ) );

		SendMessage( wnd_progress_tooltip, TTM_ADDTOOL, 0, ( LPARAM )( LPTOOLINFO )&ti_progress );


		// Add CBT hook -> detect toolbar or popup menu
		hCBT = SetWindowsHookEx( WH_CBT, &hookproc_cbt, core_api::get_my_instance(), GetCurrentThreadId() );
 	}

	virtual void on_quit()
	{
		if( g_bIntegrate ) 
		{
			// Get band visibilty/size/break
			REBARBANDINFO rbbi_get;
			ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
			rbbi_get.cbSize	= sizeof( REBARBANDINFO );
			rbbi_get.fMask	= RBBIM_STYLE | RBBIM_SIZE | RBBIM_STYLE;
	
			int iCurIndex = SendMessage( wnd_rebar, RB_IDTOINDEX, ( WPARAM )( UINT )PSEUDO_UNIQUE_BAND_ID, 0 );
			SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )iCurIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

			// Remove band (or trouble)
			SendMessage( wnd_rebar, RB_DELETEBAND, ( WPARAM )iCurIndex, 0 );
		
			// Save band settings
			g_iBandIndex	= iCurIndex;
			g_bBandShow		= !( rbbi_get.fStyle & RBBS_HIDDEN );
			g_iBandWidth	= rbbi_get.cx;
			g_bBandBreak	= ( bool )( rbbi_get.fStyle & RBBS_BREAK );
		}


		// Restore initial volume
		if( g_bResetOnQuit ) m->setVolumeRAW( dwInitVol );

		// Delete mixer
		if( m ) delete m;

		// Remove CBT hook
		if( hCBT != NULL ) UnhookWindowsHookEx( hCBT );

		// Restore wndproc REBAR
		if( ( WNDPROC )GetWindowLong( wnd_rebar, GWL_WNDPROC ) == wndproc_rebar )
		{
			SetWindowLong( wnd_rebar, GWL_WNDPROC, ( LONG )wndproc_rebar_old );
		}
	}

	virtual void on_system_shutdown()
	{
		on_quit();
	}
};



class mixer_menu : public menu_item_main
{
public:
	virtual bool get_display_data( unsigned n, const ptr_list_base<metadb_handle> & data,string_base & out, unsigned & displayflags, const GUID & caller )
	{
		if( ( n < 0 ) || ( n > get_num_items() ) || !is_available( n ) ) return false;
		
		string8 temp;
		enum_item( n, temp );

		// Find last slash
		const char * ptr = strrchr( temp, '/' );
		if( ptr )
		{
			ptr++;
		}
		else
		{
			ptr = temp;
		}

		if( ( MIXER_ACTION_FADE_STOP == n ) || ( MIXER_ACTION_FADE_CLOSE == n ) )
		{
			// Pipe to slash
			temp.replace_char( '|','/' );
		}
		out.set_string(ptr);

		displayflags = ( is_checked( n ) ? FLAG_CHECKED : 0) | ( is_disabled( n ) ? FLAG_DISABLED_GRAYED : 0 );
		return true;
	}
	
	virtual unsigned get_num_items()
	{
		return ( m ? MIXER_ACTIONS : 0 );
	}

	virtual void enum_item( unsigned n, string_base & out )
	{
		switch( n )
		{
			case MIXER_ACTION_UP_DB_01:
			{
				out = "Playback/Mixer/Volume up (+0,1 dB)";
				break;
			}
			case MIXER_ACTION_UP_DB_10:
			{
				out = "Playback/Mixer/Volume up (+1,0 dB)";
				break;
			}
			case MIXER_ACTION_UP_PERC_05:
			{
				out = "Playback/Mixer/Volume up (+5%)";
				break;
			}
			case MIXER_ACTION_UP_PERC_10:
			{
				out = "Playback/Mixer/Volume up (+10%)";
				break;
			}
			case MIXER_ACTION_DOWN_DB_01:
			{
				out = "Playback/Mixer/Volume down (-0,1 dB)";
				break;
			}
			case MIXER_ACTION_DOWN_DB_10:
			{
				out = "Playback/Mixer/Volume down (-1,0 dB)";
				break;
			}
			case MIXER_ACTION_DOWN_PERC_05:
			{
				out = "Playback/Mixer/Volume down (-5%)";
				break;
			}
			case MIXER_ACTION_DOWN_PERC_10:
			{
				out = "Playback/Mixer/Volume down (-10%)";
				break;
			}		
			case MIXER_ACTION_FADE_FULL:
			{
				out = "Playback/Mixer/Volume fade to FULL";
				break;
			}
			case MIXER_ACTION_FADE_ZERO:
			{
				out = "Playback/Mixer/Volume fade to ZERO";
				break;
			}
			case MIXER_ACTION_MUTE:
			{
				out = "Playback/Mixer/Toggle mute";
				break;
			}
			case MIXER_ACTION_RESET_INIT:
			{
				out = "Playback/Mixer/Volume RESET to initial";
				break;
			}
			case MIXER_ACTION_FADE_STOP:
			{
				out = "Playback/Stop w| fadeout";
				break;
			}
			case MIXER_ACTION_FADE_CLOSE:
			{
				out = "Foobar2000/Close w| fadeout";
				break;
			}
		}
	}

	virtual void perform_command( unsigned n )
	{
		if( !m ) return;
		if( m->getError() ) return;

	    DWORD step = 0;
		switch( n )
		{
			case MIXER_ACTION_UP_DB_01:
			case MIXER_ACTION_UP_DB_10:
			{
				DWORD dw = m->getVolumeRAW();
				if( 65535 == dw ) return;

				dw = mixer::conv_RAW_to_DB( dw );
				dw += ( n == MIXER_ACTION_UP_DB_01 ) ? 1 : 10;

				if( m->getError() )
				{
					console::error( "error getting volume" );
					console::popup();
					return;
				}

				m->setVolumeDB( dw );

				if( m->getError() )
				{
					console::info( "error setting volume" );
					console::popup();
				}

				return;
			}
			case MIXER_ACTION_UP_PERC_05:
			case MIXER_ACTION_UP_PERC_10:
			{
				DWORD dw = m->getVolumeRAW();
				if( 65535 == dw ) return;

				double f = mixer::conv_RAW_to_Percent( dw );
				f += ( n == MIXER_ACTION_UP_PERC_05 ) ? 5 : 10;

				if( m->getError() )
				{
					console::error( "error getting volume" );
					console::popup();
					return;
				}

				m->setVolumePercent( f );

				if( m->getError() )
				{
					console::info( "error setting volume" );
					console::popup();
				}

				return;
			}
			case MIXER_ACTION_DOWN_DB_01:
		    case MIXER_ACTION_DOWN_DB_10:
			{
				DWORD dw = m->getVolumeRAW();
				if( 0 == dw ) return;

				dw = mixer::conv_RAW_to_DB( dw );
				dw -= ( n == MIXER_ACTION_DOWN_DB_01 ) ? 1 : 10;

				if( m->getError() )
				{
					console::error( "error getting volume" );
					console::popup();
					return;
				}

				m->setVolumeDB( dw );

				if( m->getError() )
				{
					console::info( "error setting volume" );
					console::popup();
				}

				return;
			}
			case MIXER_ACTION_DOWN_PERC_05:
		    case MIXER_ACTION_DOWN_PERC_10:
			{
				DWORD dw = m->getVolumeRAW();
				if( 0 == dw ) return;

				double f = mixer::conv_RAW_to_Percent( dw );
				f -= ( n == MIXER_ACTION_DOWN_PERC_05 ) ? 5 : 10;

				if( m->getError() )
				{
					console::error( "error getting volume" );
					console::popup();
					return;
				}

				m->setVolumePercent( f );

				if( m->getError() )
				{
					console::info( "error setting volume" );
					console::popup();
				}

				return;
			}
			case MIXER_ACTION_FADE_FULL:
			{
				if( ( FADE::bActive ) && ( FADE::dir == FADE_FULL ) ) return;
				FADE::doFade( FADE_FULL, false, false );

				return;
			}
			case MIXER_ACTION_FADE_ZERO:
			{
				if( ( FADE::bActive ) && ( FADE::dir == FADE_ZERO ) ) return;
				FADE::doFade( FADE_ZERO, false, false );

				return;
			}
			case MIXER_ACTION_MUTE:
			{
				
				bool bNow = m->getMute();

				if( m->getError() )
				{
					console::error( "error getting mute" );
					console::popup();
					return;
				}

				m->setMute( !bNow );

				if( m->getError() )
				{
					console::error( "error setting mute" );
					console::popup();
				}

				return;
			}
			case MIXER_ACTION_RESET_INIT:
			{
				DWORD dw = m->getVolumeRAW();
				if( dwInitVol == dw ) return;

				if( m->getError() )
				{
					console::error( "error getting volume" );
					console::popup();
					return;
				}

				m->setVolumeRAW( dwInitVol );

				if( m->getError() )
				{
					console::info( "error setting volume" );
					console::popup();
				}

				return;
			}
			case MIXER_ACTION_FADE_STOP:
			{
				play_control * pc = play_control::get();
				if( !pc->is_playing() || ( pc->is_playing() && pc->is_paused() ) ) return;

				FADE::doFade( FADE_ZERO, true, false );

				return;
			}
			case MIXER_ACTION_FADE_CLOSE:
			{
				play_control * pc = play_control::get();
				if( !pc->is_playing() || ( pc->is_playing() && pc->is_paused() ) )
				{
					// Close without fade
					PostMessage( core_api::get_main_window(), WM_CLOSE, 0, 0 );
				}
				else
				{
					FADE::doFade( FADE_ZERO, true, true );
				}

				return;
			}
		}
	}

	virtual bool is_checked( unsigned n )
	{
		switch( n )
		{
			case MIXER_ACTION_MUTE:
			{
				if( !m ) return false;
				return !m->getMute();		
			}
		}
		return false;
	}

	virtual bool is_disabled( unsigned n )
	{
		switch( n )
		{
			case MIXER_ACTION_UP_DB_01:
			case MIXER_ACTION_UP_DB_10:
			case MIXER_ACTION_UP_PERC_05:
			case MIXER_ACTION_UP_PERC_10:
			{
				return ( 65535 == m->getVolumeRAW() );
			}
			case MIXER_ACTION_FADE_FULL:
			{
				return ( FADE::bActive && ( FADE::dir == FADE_FULL ) ) || ( 65535 == m->getVolumeRAW() );
			}
			case MIXER_ACTION_DOWN_DB_01:
			case MIXER_ACTION_DOWN_DB_10:
			case MIXER_ACTION_DOWN_PERC_05:
			case MIXER_ACTION_DOWN_PERC_10:
			{
				return ( 0 == m->getVolumeRAW() );
			}
			case MIXER_ACTION_FADE_ZERO:
			{
				return ( FADE::bActive && ( FADE::dir == FADE_ZERO ) ) || ( 0 == m->getVolumeRAW() );
			}
			case MIXER_ACTION_RESET_INIT:
			{
				return ( m->getVolumeRAW() == dwInitVol );
			}
			case MIXER_ACTION_FADE_STOP:
			{
				play_control * pc = play_control::get();
				return ( !pc->is_playing() || ( pc->is_playing() && pc->is_paused() ) );
			}
			// case MIXER_ACTION_FADE_CLOSE: ALWAYS ENABLED
		}
		return false;
	}

	virtual bool is_available( unsigned n )
	{
		return !( m->getError() );
	}
};



BOOL CALLBACK wndproc_fade( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	switch( message )
	{
		case WM_KEYDOWN:
		{
			switch( wp )
			{
				case VK_ADD:
				case VK_PRIOR:		// page up
				{
					BOOL succ = FALSE;
					UINT res = 0;

					if( hwnd == wnd_fadein )
					{
						res = GetDlgItemInt( wnd_config, IDC_FADEINTIME, &succ, FALSE );
					}
					else if( hwnd == wnd_fadeout )
					{
						res = GetDlgItemInt( wnd_config, IDC_FADEOUTTIME, &succ, FALSE );
					}

					if( succ )
					{
						if( res == FADE_TIME_MAX )
						{
							break;
						}
						else if( res > FADE_TIME_MAX )
						{
							res = FADE_TIME_MAX;
						}
						else
						{
							if( VK_ADD == wp )
							{
								res++;
							}
							else
							{
								// 49 -> 50
								// 53 -> 60
								res = ( ( res / 10 ) + 1 ) * 10;
							}
						}

						if( hwnd == wnd_fadein )
						{
							SetDlgItemInt( wnd_config, IDC_FADEINTIME, res, FALSE );
						}
						else
						{
							SetDlgItemInt( wnd_config, IDC_FADEOUTTIME, res, FALSE );
						}
					}
					break;
				}
				case VK_SUBTRACT:
				case VK_NEXT:		// page down
				{
					BOOL succ = FALSE;
					UINT res = 0;

					if( hwnd == wnd_fadein )
					{
						res = GetDlgItemInt( wnd_config, IDC_FADEINTIME, &succ, FALSE );
					}
					else if( hwnd == wnd_fadeout )
					{
						res = GetDlgItemInt( wnd_config, IDC_FADEOUTTIME, &succ, FALSE );
					}

					if( succ )
					{
						if( res == FADE_TIME_MIN )
						{
							break;
						}
						else if( res < FADE_TIME_MIN )
						{
							res = FADE_TIME_MIN;
						}
						else
						{
							if( VK_SUBTRACT == wp )
							{
								res--;
							}
							else
							{
								// 41 -> 40
								// 40 -> 30
								res = ( ( res / 10 ) - ( ( res % 10 ) ? 0 : 1 ) ) * 10;
							}
						}	

						if( hwnd == wnd_fadein )
						{
							SetDlgItemInt( wnd_config, IDC_FADEINTIME, res, FALSE );
						}
						else
						{
							SetDlgItemInt( wnd_config, IDC_FADEOUTTIME, res, FALSE );
						}
					}
					break;
				}
			}
			break;
		}
	}

	if( hwnd == wnd_fadein )
	{
		return CallWindowProc( wndproc_fadein_old, hwnd, message, wp, lp );
	}
	else
	{
		return CallWindowProc( wndproc_fadeout_old, hwnd, message, wp, lp );
	}
}

BOOL CALLBACK wndproc_config( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	static HWND wnd_frame		= NULL;

	static bool bIntegrate_OLD			= false;
	static bool bBandLabel_OLD			= false;
	static bool bSetGripperAlways_OLD	= false;
	static bool bDSPInstead_OLD			= false;

	switch( message )
	{
		case WM_INITDIALOG:
		{
			wnd_config = hwnd;
			bIntegrate_OLD			= g_bIntegrate;
			bBandLabel_OLD			= g_bBandLabel;
			bSetGripperAlways_OLD	= g_bGripperAlways;
			bDSPInstead_OLD			= g_bDSPInstead;

			// Get control handles
			wnd_fadein		= GetDlgItem( hwnd, IDC_FADEINTIME	);
			wnd_fadeout		= GetDlgItem( hwnd, IDC_FADEOUTTIME	);
			wnd_frame		= GetDlgItem( hwnd, IDC_FRAME	);

			// Apply current settings
			CheckDlgButton(	hwnd, IDC_INTEGRATE, g_bIntegrate );
			CheckRadioButton( hwnd, IDC_MAIN, IDC_DSP, ( g_bDSPInstead ? IDC_DSP : IDC_MAIN ) );
			SetDlgItemInt( hwnd, IDC_FADEINTIME, g_iFadeTimeFull, FALSE );
			SetDlgItemInt( hwnd, IDC_FADEOUTTIME, g_iFadeTimeZero, FALSE );
			CheckDlgButton(	hwnd, IDC_RESTOREONQUIT, g_bResetOnQuit );
			CheckDlgButton( hwnd, IDC_GRIPPERALWAYS, g_bGripperAlways );
			CheckDlgButton( hwnd, IDC_LABEL, g_bBandLabel );

			// Exchance wndproc FADEIN
			wndproc_fadein_old = ( WNDPROC )GetWindowLong( wnd_fadein, GWL_WNDPROC );
			if( wndproc_fadein_old != NULL )
			{
				SetWindowLong( wnd_fadein, GWL_WNDPROC, ( LONG )wndproc_fade );
			}

			// Exchance wndproc FADEOUT
			wndproc_fadeout_old = ( WNDPROC )GetWindowLong( wnd_fadeout, GWL_WNDPROC );
			if( wndproc_fadeout_old != NULL )
			{
				SetWindowLong( wnd_fadeout, GWL_WNDPROC, ( LONG )wndproc_fade );
			}

			break;
		}
		case WM_SIZE:
		{
			// Resize frame
			RECT r;
			GetWindowRect( hwnd, &r );
			MoveWindow( wnd_frame, 4, 0, r.right - r.left - 4, r.bottom - r.top - 5, TRUE );

			break;
		}
		case WM_DESTROY:
		{
			// Save settings
			g_bIntegrate = IsDlgButtonChecked( hwnd, IDC_INTEGRATE );
			g_bDSPInstead = IsDlgButtonChecked( hwnd, IDC_DSP );

			BOOL succ = FALSE;
			UINT res = 0;
			res = GetDlgItemInt( hwnd, IDC_FADEINTIME, &succ, FALSE );
			if( succ ) g_iFadeTimeFull = __max( FADE_TIME_MIN, __min( FADE_TIME_MAX, res ) );
			res = GetDlgItemInt( hwnd, IDC_FADEOUTTIME, &succ, FALSE );
			if( succ ) g_iFadeTimeZero = __max( FADE_TIME_MIN, __min( FADE_TIME_MAX, res ) );

			g_bResetOnQuit    = IsDlgButtonChecked( hwnd, IDC_RESTOREONQUIT );
			g_bGripperAlways  = IsDlgButtonChecked( hwnd, IDC_GRIPPERALWAYS );
			g_bBandLabel      = IsDlgButtonChecked( hwnd, IDC_LABEL );


			// Apply changes
			if( !bWaitingRebar )
			{
				bool bBandAdded = false;
				int iBandCount = -1;

				REBARBANDINFO rbbi_vol;
				ZeroMemory( &rbbi_vol, sizeof( REBARBANDINFO ) );
				rbbi_vol.cbSize			= sizeof( REBARBANDINFO );
				rbbi_vol.fMask			= RBBIM_CHILD | RBBIM_CHILDSIZE | RBBIM_ID | RBBIM_STYLE |
											( g_bBandLabel ? RBBIM_TEXT : 0 ) |
											( ( g_iBandWidth != -1 ) ? RBBIM_SIZE : 0 );
				rbbi_vol.hwndChild		= wnd_progress;
				rbbi_vol.lpText			= "Volume";
				rbbi_vol.cxMinChild		= 113;
				rbbi_vol.cyMinChild		= 18;
				rbbi_vol.cyChild		= 18;
				rbbi_vol.cyMaxChild 	= 18;
				rbbi_vol.wID			= PSEUDO_UNIQUE_BAND_ID;
				rbbi_vol.cx				= g_iBandWidth;
				rbbi_vol.fStyle			= RBBS_CHILDEDGE | ( g_bGripperAlways ? RBBS_GRIPPERALWAYS : 0 ) |
											( ( g_bBandShow ) ? 0 : RBBS_HIDDEN ) |
											( ( g_bBandBreak ) ? RBBS_BREAK : 0 );

				REBARBANDINFO rbbi_get;
				ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
				rbbi_get.cbSize	= sizeof( REBARBANDINFO );
				rbbi_get.fMask	= RBBIM_STYLE;

				
				if( g_bIntegrate != bIntegrate_OLD )
				{
					if( g_bIntegrate )
					{

					
/******************************************************************************************************
**                      Add volume band                                                              **
******************************************************************************************************/

						if( g_iBandIndex > ( iBandCount = SendMessage( wnd_rebar, RB_GETBANDCOUNT, 0, 0 ) ) )
						{
							g_iBandIndex = iBandCount;
						}

						SendMessage( wnd_rebar, RB_INSERTBAND, ( WPARAM )g_iBandIndex, ( LPARAM )( LPREBARBANDINFO )&rbbi_vol );

						if( !g_bBandShow ) ShowWindow( wnd_progress, SW_HIDE );



						bBandAdded = true;
					}
					else
					{

					
/******************************************************************************************************
**                      Remove volume band                                                              **
******************************************************************************************************/

						// Get band visibilty/size/break
						REBARBANDINFO rbbi_get;
						ZeroMemory( &rbbi_get, sizeof( REBARBANDINFO ) );
						rbbi_get.cbSize	= sizeof( REBARBANDINFO );
						rbbi_get.fMask	= RBBIM_STYLE | RBBIM_SIZE | RBBIM_STYLE;
	
						int iCurIndex = SendMessage( wnd_rebar, RB_IDTOINDEX, ( WPARAM )( UINT )PSEUDO_UNIQUE_BAND_ID, 0 );
						SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )iCurIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

						// Remove band (or trouble)
						SendMessage( wnd_rebar, RB_DELETEBAND, ( WPARAM )iCurIndex, 0 );
		
						// Save band settings
						g_iBandIndex	= iCurIndex;
						g_bBandShow		= !( rbbi_get.fStyle & RBBS_HIDDEN );
						g_iBandWidth	= rbbi_get.cx;
						g_bBandBreak	= ( bool )( rbbi_get.fStyle & RBBS_BREAK );
					}
				}


/******************************************************************************************************
**              Change gripper settings                                                              **
******************************************************************************************************/

				if
				(
					g_bIntegrate && 
					(
						( ( bool )g_bIntegrate		!= bIntegrate_OLD ) ||
						( ( bool )g_bGripperAlways	!= bSetGripperAlways_OLD )
					)
				)
				{
					if( !bBandAdded )
					{
						// Subtracting 1 we can use the exact code from [on_init]...
						iBandCount = -1 + SendMessage( wnd_rebar, RB_GETBANDCOUNT, 0, 0 );
					}


/******************************************************************************************************
**                  RBBS_GRIPPERALWAYS on/off + check for locked/RBBS_NOGRIPPER                      **
******************************************************************************************************/

					bool bNoGripper = false;
					iBandCount++;
					for( int i = 0; i < iBandCount; i++ )
					{
						// Exclude volume band
						if( i == g_iBandIndex ) continue;

						// Get band info
						SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )i, ( LPARAM )( REBARBANDINFO * )&rbbi_get );

						// RBBS_GRIPPERALWAYS: Add where missing / remove where set
						if( ( bool )g_bGripperAlways ^ ( bool )( rbbi_get.fStyle & RBBS_GRIPPERALWAYS ) )		// xor
						{
							if( g_bGripperAlways )
							{
								rbbi_get.fStyle	|= RBBS_GRIPPERALWAYS;
							}
							else
							{
								rbbi_get.fStyle	-= RBBS_GRIPPERALWAYS;
							}

							// Only the values specified in [.fMask] are overwritten???
							SendMessage( wnd_rebar, RB_SETBANDINFO, ( WPARAM )( UINT )i, ( LPARAM )( REBARBANDINFO * )&rbbi_get );
						}

						// Toolbar locked?
						bNoGripper |= ( rbbi_get.fStyle & RBBS_NOGRIPPER );
					}


					// Update volume band gripper
					REBARBANDINFO * lprbbi_vol = NULL;
					if( bBandAdded )
					{
						lprbbi_vol = &rbbi_vol;
					}
					else
					{
						// Get band info
						SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )g_iBandIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_get );
						lprbbi_vol = &rbbi_get;
					}

					if( bNoGripper )
					{
						lprbbi_vol->fStyle |= RBBS_NOGRIPPER;
					}

					bool bChanged = false;
					if( g_bGripperAlways )
					{
						lprbbi_vol->fStyle |= RBBS_GRIPPERALWAYS;
						bChanged = true;
					}
					else
					{
						if( lprbbi_vol->fStyle & RBBS_GRIPPERALWAYS )
						{
							lprbbi_vol->fStyle -= RBBS_GRIPPERALWAYS;
							bChanged = true;
						}
					}

					// Also do label settings
					if( ( bool )g_bBandLabel != bBandLabel_OLD )
					{
						lprbbi_vol->fMask |= RBBIM_TEXT;
						if( g_bBandLabel )
						{
							lprbbi_vol->lpText	= "Volume";
							lprbbi_vol->cch		= 6;
						}
						else
						{
							lprbbi_vol->lpText	= "";
							lprbbi_vol->cch		= 0;
						}

						bChanged = true;
					}

					if( bNoGripper || bChanged )
					{
						SendMessage( wnd_rebar, RB_SETBANDINFO, ( WPARAM )( UINT )g_iBandIndex, ( LPARAM )( REBARBANDINFO * )lprbbi_vol );
					}
				}



/******************************************************************************************************
**              Change label settings                                                              **
******************************************************************************************************/

				if
				(
					g_bIntegrate && 
					( ( bool )g_bGripperAlways	== bSetGripperAlways_OLD ) &&
					( ( bool )g_bBandLabel		!= bBandLabel_OLD )
				)
				{
					// Grippers have not been updated so
					// label has not changed before

					// Update volume band label
					REBARBANDINFO * lprbbi_vol = NULL;
					if( bBandAdded )
					{
						lprbbi_vol = &rbbi_vol;
					}
					else
					{
						// Get band info
						SendMessage( wnd_rebar, RB_GETBANDINFO, ( WPARAM )( UINT )g_iBandIndex, ( LPARAM )( REBARBANDINFO * )&rbbi_get );
						lprbbi_vol = &rbbi_get;
					}

					lprbbi_vol->fMask |= RBBIM_TEXT;
					if( g_bBandLabel )
					{
						lprbbi_vol->lpText	= "Volume";
						lprbbi_vol->cch		= 6;
					}
					else
					{
						lprbbi_vol->lpText	= "";
						lprbbi_vol->cch		= 0;
					}

					// Only the values specified in [.fMask] are overwritten???
					SendMessage( wnd_rebar, RB_SETBANDINFO, ( WPARAM )( UINT )g_iBandIndex, ( LPARAM )( REBARBANDINFO * )lprbbi_vol );
				}



/******************************************************************************************************
**              Apply new target to progress bar                                                     **
******************************************************************************************************/

				if
				(
					g_bIntegrate &&
					( g_bIntegrate != bIntegrate_OLD ) &&
					( g_bDSPInstead != bDSPInstead_OLD )
				)
				{
					if( g_bDSPInstead )
					{
						int iVal = ( int )( ( play_control::get()->get_volume() - iN ) / fM ) + 1;
						iVal = __max( 0, __min( iVal, 65535 ) );
						PostMessage( wnd_progress, PBM_SETPOS , ( WPARAM )iVal, 0 );
					}
					else
					{
						SendMessage( wnd_progress, PBM_SETPOS , ( WPARAM )m->getVolumeRAW(), 0 );
					}
				}
			}

			break;
		}
	}
	return 0;
}


class mixer_config : public config
{
public:

	virtual HWND create( HWND parent )
	{
		return CreateDialog( core_api::get_my_instance(), MAKEINTRESOURCE( IDD_CONFIG ), parent, wndproc_config );
	}

	virtual const char * get_name()
	{
		return "Direct Volume Control";
	}

	virtual const char * get_parent_name()
	{
		return "Components";
	}
};


class mixer_callback : public play_callback
{
public:
	virtual void on_playback_starting() { }
	virtual void on_playback_new_track(metadb_handle * track) { }
	virtual void on_playback_stop(play_control::stop_reason) { }
	virtual void on_playback_seek(double time) { }
	virtual void on_playback_pause(int state) { }
	virtual void on_playback_edited(metadb_handle * track) { }
	virtual void on_playback_dynamic_info(const file_info * info,bool b_track_change) { }
	virtual void on_volume_change( int new_val )
	{
		if( !bWaitingRebar && g_bDSPInstead && !bDown )
		{
			int iVal = ( int )( ( new_val - iN ) / fM );
			iVal = __max( 0, __min( iVal, 65535 ) );
			PostMessage( wnd_progress, PBM_SETPOS , ( WPARAM )iVal, 0 );
		}
	}
};


static service_factory_single_t< initquit,		mixer_init		> foo_init;
static service_factory_single_t< menu_item,		mixer_menu		> foo_menu;
static service_factory_single_t< config,		mixer_config	> foo_config;
static service_factory_single_t< play_callback,	mixer_callback	> foo_callback;


DECLARE_COMPONENT_VERSION
(
	"Direct Volume Control",
	"0.96 BETA // 2004-09-08",
	"Now stop the whining!\n\n"
	"Written by...\n"
	"* Messer\t\thttp://www.dago.pmp.com.pl  \n"
	"* hartwork\thttp://www.hartwork.org  "
);
